package example.hearts.ai;

import java.io.*;

import inonit.debug.*;

import inonit.domain.game.card.*;
import inonit.domain.game.card.hearts.*;
import inonit.domain.game.card.hearts.ai.*;

public class Izzy extends AbstractIntelligence {
	private Rank.Order rankOrder = Hearts.get().getRankOrder();
	private Sort.Order DESCENDING;
	private Sort.Order ASCENDING;

	public Izzy() {
		DESCENDING = Sort.compose( new Sort.Order[] { Hearts.get().getRankOrder().getDescendingOrder(), new SuitOrder() } );
		ASCENDING = Sort.reverse(DESCENDING);
	}

	public String getName() {
		return "Izzy";
	}

	public void passWasIllegal(String s) {
		println(s);
	}

	public void playWasIllegal(String s) {
		println(s);
	}

	private CardArray hand;
	private CardArray pass;
	private int leftToPass;

	public CardArray getPass(int numberToPass) {
		// check for void suit
		// check for can go void
		// check for almost void - 1 or 2 left in suit
		// check spades - queen verify
		// check for lots of one suit
		// check for low number of high hearts
		// check for low number of high cards in any suit

		leftToPass = numberToPass;
		hand = getHandX();
		pass = new CardArray();

		println("Current hand:\n" + cardListBySuit(hand));

		tryToGoVoid(Suit.SPADES);
		tryToGoVoid(Suit.CLUBS);
		tryToGoVoid(Suit.HEARTS);
		tryToGoVoid(Suit.DIAMONDS);

		passHighOf(Suit.SPADES);
		passHighOf(Suit.HEARTS);

		removeFromHand(Suit.DIAMONDS);
		passHighestLeft();

		println("Pass list: " + cardList(pass) + "\n");
		return pass;
	}

	private boolean isFinished() {
		return leftToPass <= 0;
	}

	private void pass(Suit suit) {
		pass(hand.get(suit));
	}

	private void pass(CardArray cards) {
		for (int i = 0; i < cards.getLength(); i++)
			pass(cards.get(i));
	}

	private void pass(Card card) {
		hand.remove(card);
		pass.add(card);
		leftToPass--;
	}


	private void tryToGoVoid(Suit suit) {
		if (couldGoVoid(suit))
			pass(suit);
		println("try to go void in " + suit.getName() + ": " + cardList(pass));
	}

	private boolean couldGoVoid(Suit suit) {
		return hand.get(suit).getLength() <= leftToPass;
	}

	private void passHighOf(Suit suit) {
		CardArray ca = hand.get(suit).sort(DESCENDING);
		int i = 0;
		while (!isFinished() && (i < ca.getLength())) {
			Card card = ca.get(i);
			//	Choose arbitrary suit to get comparison
			if (isHigher(card, Card.TEN_CLUBS))
				pass(card);
			i++;
		}
		println("pass high of " + suit.getName() + ": " + cardList(pass));
	}

	private void removeFromHand(Suit suit) {
		hand.remove(hand.get(suit));
	}

	private void passHighestLeft() {
		hand = hand.sort(DESCENDING);
		while (!isFinished())
			pass(hand.get(0));
		println("pass highest left: " + cardList(pass));
	}

	private CardArray getHandX() {
		return getCards();
	}

	private CardArray getCurrentTrickCards() {
		return getTrick().getCards();
	}

	private Suit getCurrentTrickSuit() {
		if (!getTrick().getCards().isEmpty())
			return getTrick().getCards().get(0).getSuit();
		return null;
	}

	private Card getCurrentTrickWinner() {
		if ( getTrick().getCards().isEmpty() ) {
			return null;
		} else {
			Card led = getTrick().getCards().get(0);
			Card leader = led;
			for (int i=1; i<getTrick().getCards().getLength(); i++) {
				if (isSameSuit(led, getTrick().getCards().get(i))) {
					if (isHigher(getTrick().getCards().get(i), leader)) {
						leader = getTrick().getCards().get(i);
					}
				}
			}
			return leader;
		}
	}
	
private int round = 0;

	public Card getPlay() {
		println("Current hand:\n" + cardListBySuit(getHandX()));
		println("Current trick [" + ((getCurrentTrickSuit() != null) ? getCurrentTrickSuit().getName() : "none") +
			"]: " + cardList(getTrick().getCards()));
		println("Current trick winner: " + getCurrentTrickWinner());
		println("Legal plays: " + cardList(getLegalPlays()));

		Card card = selectCard();
		if (card == null) {
			println("NO CARD SELECTED!");
			card = getLegalPlays().get(0);
		}

		println("Selected: " + card);
		println("");

round++;
//if (round == 2)
//	exit();

		return card;
	}

	private boolean isVoidedInTrickSuit() {
		return getCards().get( getTrick().getCards().get(0).getSuit() ).getLength() == 0;
	}
	
	private Card selectCard() {
		if (getLegalPlays().getLength() == 1)
			return getLegalPlays().get(0);

		if (getTricks().length == 0) {
			if (canPlay(Suit.CLUBS))
				return playHighest();

			if (canPlay(Suit.SPADES))
				return playHighest(Suit.SPADES);

			return playHighest();
		}

		if (getTrick().getLed() == 0) {
			if (canPlay(Suit.SPADES)) {
				CardArray legals = getLegalPlays().get(Suit.SPADES).sort(DESCENDING);
				if (!legals.contains(Card.ACE_SPADES) && !legals.contains(Card.KING_SPADES) && !legals.contains(Card.QUEEN_SPADES))
					return legals.get(0);
			}

			if (canPlay(Suit.DIAMONDS))
				return playLowest(Suit.DIAMONDS);
			if (canPlay(Suit.CLUBS))
				return playLowest(Suit.CLUBS);

			return playLowest();
		}

		if (isVoidedInTrickSuit()) {
			Suit suit = getAlmostVoid();
			if (suit != null)
				return playHighest(suit);

			if (canPlay(Suit.SPADES)) {
				if (canPlay(Card.QUEEN_SPADES))
					return Card.QUEEN_SPADES;
				if (canPlay(Card.ACE_SPADES))
					return Card.ACE_SPADES;
				if (canPlay(Card.KING_SPADES))
					return Card.KING_SPADES;
			}

			if (canPlay(Suit.HEARTS))
				return playHighest(Suit.HEARTS);

			return playHighest();
		}

		if (getTrick().getLed() == 1) {
			if (!getCurrentTrickSuit().equals(Suit.DIAMONDS)) {
				Card card = getHighestNoTake();
				Card winner = getCurrentTrickWinner();
				if (isHigher(card, winner))
					return playHighest();
				else
					return card;
			}
		}

		if (getCurrentTrickSuit().equals(Suit.DIAMONDS)) {
			if (getTrick().getCards().contains(Card.JACK_DIAMONDS) && !getTrick().getCards().contains(Card.QUEEN_SPADES))
				return playHighest();
		}

		return getHighestNoTake();
//		// if null, get highest
//		return playLowest();
	}


	private boolean canPlay(Suit suit) {
		return contains(getLegalPlays(), suit);
	}

	private boolean canPlay(Card card) {
		return getLegalPlays().contains(card);
	}

	private Card playLowest() {
		return getLowest(getLegalPlays());
	}

	private Card playLowest(Suit suit) {
		return getLowest(getLegalPlays().get(suit));
	}

	private Card playHighest() {
		return getHighest(getLegalPlays());
	}

	private Card playHighest(Suit suit) {
		return getHighest(getLegalPlays().get(suit));
	}


	private Suit getAlmostVoid() {
		if (isAlmostVoid(Suit.SPADES))
			return Suit.SPADES;
		if (isAlmostVoid(Suit.HEARTS))
			return Suit.HEARTS;
		if (isAlmostVoid(Suit.DIAMONDS))
			return Suit.DIAMONDS;
		if (isAlmostVoid(Suit.CLUBS))
			return Suit.CLUBS;
		return null;
	}

	private boolean isAlmostVoid(Suit suit) {
		return getLegalPlays().getCount(Suit.SPADES) == 1;
	}


	private boolean contains(CardArray cards, Suit suit) {
		return cards.getCount(suit) > 0;
	}

	private Card getLowest(CardArray cards) {
		return cards.sort(ASCENDING).get(0);
	}

	private Card getHighest(CardArray cards) {
		return cards.sort(DESCENDING).get(0);
	}


//public Card getPlay() {
//	if (getSituation().isNextPlayLead()) {
//		return getHighest();
//	} else {
//		if (canTakeTrick()) {
//			if (getSituation().isNextPlayLast()) {
//				return getLowestToTake();
//			} else {
//				return getHighest();
//			}
//		} else {
//			return getLowest();
//		}
//	}
//}


//private boolean canTakeTrick() {
//	boolean rv = false;
//	Card currentWinner = getSituation().getCurrentTrickWinner();
//	Card highestPlayInSuit = getHighest();
//	if (highestPlayInSuit != null && currentWinner != null) {
//		rv = rankOrder.firstIsHigher(highestPlayInSuit.getRank(), currentWinner.getRank());
//	}
//	return rv;
//}

	private Card getLowestToTake() {
		Card currentWinner = getCurrentTrickWinner();
		Card[] legalPlays = getLegalPlays().sort(DESCENDING).get();
		Card rv = legalPlays[0];
		for (int i = 0; i < legalPlays.length; i++) {
			if (isHigher(legalPlays[i], currentWinner)) {
				rv = legalPlays[i];
			}
		}
		return rv;
	}

	private Card getHighestNoTake() {
		Card currentWinner = getCurrentTrickWinner();
		Card[] legalPlays = getLegalPlays().sort(ASCENDING).get();
		Card rv = legalPlays[0];
		for (int i = 0; i < legalPlays.length; i++) {
			if (!isHigher(legalPlays[i], currentWinner)) {
				rv = legalPlays[i];
			}
		}
		return rv;
	}

//private Card getHighest() {
//	return getSituation().getLegalPlays().sort(DESCENDING).get(0);
//}

//private Card getLowest() {
//	CardArray hearts = getSituation().getLegalPlays().get(Suit.HEARTS);
//	CardArray legalPlays = getSituation().getLegalPlays();
//	legalPlays.remove(hearts);
//	if (legalPlays.contains(Card.QUEEN_SPADES))
//		legalPlays.remove(Card.QUEEN_SPADES);
//	if (legalPlays.contains(Card.JACK_DIAMONDS))
//		legalPlays.remove(Card.JACK_DIAMONDS);
//	CardArray sorted = legalPlays.sort(ASCENDING);
//	if (sorted.get().length != 0) {
//		return sorted.get(0);
//	} else {
//		return getSituation().getLegalPlays().get(0);
//	}
//}


	public static String cardList(CardArray cards) {
		String rv = "";
		if (cards != null) {
			for (int i = 0; i < cards.getLength(); i++)
				rv += cards.get(i).toString() + " ";
		}
		return rv;
	}

	public static String cardListBySuit(CardArray cards) {
		String rv = "";
		rv += cardList(cards.get(Suit.CLUBS)) + "\n";
		rv += cardList(cards.get(Suit.DIAMONDS)) + "\n";
		rv += cardList(cards.get(Suit.SPADES)) + "\n";
		rv += cardList(cards.get(Suit.HEARTS));
		return rv;
	}

	private void println(Object o) {
		log(o);
	}

	public static class SuitOrder extends Card.Order {
		private int getSuitValue(Suit suit) {
			if (suit.equals(Suit.SPADES))
				return 3;
			else if (suit.equals(Suit.HEARTS))
				return 4;
			else if (suit.equals(Suit.CLUBS))
				return 1;
			else if (suit.equals(Suit.DIAMONDS))
				return 2;
			else {
				Debug.crash("Unrecognized suit: " + suit.getName());
				return 0;
			}
		}

		public Sort.Action compare(Card one, Card two) {
			if (getSuitValue(one.getSuit()) > getSuitValue(two.getSuit()))
				return Sort.Action.LEAVE;
			else if (getSuitValue(two.getSuit()) > getSuitValue(one.getSuit()))
				return Sort.Action.SWAP;
			else
				return Sort.Action.EQUAL;
		}
	}
}
