package example.hearts.ai;

import java.io.*;

import inonit.debug.*;

import inonit.domain.game.card.*;
import inonit.domain.game.card.hearts.*;
import inonit.domain.game.card.hearts.ai.*;

public class Kamikaze extends AbstractIntelligence {
	private Rank.Order rankOrder = Hearts.get().getRankOrder();
	private Sort.Order DESCENDING;
	private Sort.Order ASCENDING;

	public Kamikaze() {
		DESCENDING = Sort.compose( new Sort.Order[] { Hearts.get().getRankOrder().getDescendingOrder(), new SuitOrder() } );
		ASCENDING = Sort.reverse(DESCENDING);
	}
	
	public String getName() {
		return "Kamikaze";
	}
	
	public void passWasIllegal(String s) {
		System.out.println(s);
	}
	
	public void playWasIllegal(String s) {
		System.out.println(s);
	}
	
	public CardArray getPass(int numberToPass) {
		CardArray cards = getCards().sort(ASCENDING);
		CardArray rv = new CardArray();
		for (int i = 0; i < numberToPass; i++) {
			rv.add(cards.get(i));
		}
		return rv;
	}
	
	private Card getCurrentTrickWinner() {
		if ( getTrick().getCards().isEmpty() ) {
			return null;
		} else {
			Card led = getTrick().getCards().get(0);
			Card leader = led;
			for (int i=1; i<getTrick().getCards().getLength(); i++) {
				if (isSameSuit(led, getTrick().getCards().get(i))) {
					if (isHigher(getTrick().getCards().get(i), leader)) {
						leader = getTrick().getCards().get(i);
					}
				}
			}
			return leader;
		}
	}
	
	private boolean canTakeTrick() {
		boolean rv = false;
		Card currentWinner = getCurrentTrickWinner();
		Card highestPlayInSuit = getHighest();
		if (highestPlayInSuit != null && currentWinner != null) {
			rv = isHigher(highestPlayInSuit, currentWinner);
		}
		return rv;
	}
	
	private Card getLowestToTake() {
		Card currentWinner = getCurrentTrickWinner();
		Card[] legalPlays = getLegalPlays().sort(DESCENDING).get();
		Card rv = legalPlays[0];
		for (int i = 0; i < legalPlays.length; i++) {
			if (isHigher(legalPlays[i], currentWinner)) {
				rv = legalPlays[i];
			}
		}
		return rv;
	}
	
	private Card getHighestNoTake() {
		Card currentWinner = getCurrentTrickWinner();
		Card[] legalPlays = getLegalPlays().sort(ASCENDING).get();
		Card rv = legalPlays[0];
		for (int i = 0; i < legalPlays.length; i++) {
			if (!isHigher(legalPlays[i], currentWinner)) {
				rv = legalPlays[i];
			}
		}
		return rv;
	}
	
	private Card getHighest() {
		return getLegalPlays().sort(DESCENDING).get(0);
	}
	
	private Card getLowest() {
		CardArray hearts = getLegalPlays().get(Suit.HEARTS);
		CardArray legalPlays = getLegalPlays();
		legalPlays.remove(hearts);
		if (legalPlays.contains(Card.QUEEN_SPADES))
			legalPlays.remove(Card.QUEEN_SPADES);
		if (legalPlays.contains(Card.JACK_DIAMONDS))
			legalPlays.remove(Card.JACK_DIAMONDS);	
		CardArray sorted = legalPlays.sort(ASCENDING);
		if (sorted.get().length != 0) {
			return sorted.get(0);
		} else {
			return getLegalPlays().get(0);
		}
	}
	
	public Card getPlay() {
		if (getTrick().getLed() == 0) {
			return getHighest();
		} else {
			if (canTakeTrick()) {
				if (getTrick().getLed() == 1) {
					return getLowestToTake();
				} else {
					return getHighest();
				}
			} else {
				return getLowest();
			}
		}
	}
		
	

	private static class SuitOrder extends Card.Order {
		private int getSuitValue(Suit suit) {
			if (suit.equals(Suit.SPADES))
				return 2;
			else if (suit.equals(Suit.HEARTS))
				return 4;
			else if (suit.equals(Suit.CLUBS))
				return 1;
			else if (suit.equals(Suit.DIAMONDS))
				return 3;
			else {
				Debug.crash("Unrecognized suit: " + suit.getName());
				return 0;
			}
		}
	
		public Sort.Action compare(Card one, Card two) {
			if (getSuitValue(one.getSuit()) > getSuitValue(two.getSuit()))
				return Sort.Action.LEAVE;
			else if (getSuitValue(two.getSuit()) > getSuitValue(one.getSuit()))
				return Sort.Action.SWAP;
			else
				return Sort.Action.EQUAL;
		}
	}
}
