package example.hearts.ai;

import java.io.*;

import inonit.debug.*;

import inonit.domain.game.card.*;
import inonit.domain.game.card.hearts.*;
import inonit.domain.game.card.hearts.ai.*;

public class Neanderthal extends AbstractIntelligence {
	//	[BUG] -- does not play queen that cannot take trick -- "have alternative"
	private Rank.Order rankOrder = Hearts.get().getRankOrder();
	private Sort.Order ASCENDING;
	private Sort.Order DESCENDING;
	
	public Neanderthal() {
		DESCENDING = Sort.compose( 
			new Sort.Order[] { 
				new SpadeOrder(),
				Hearts.get().getRankOrder().getDescendingOrder(), 
				new SuitOrder() 
			} 
		);
		ASCENDING = Sort.reverse(DESCENDING);
	}
	
	public String getName() {
		return "Neanderthal";
	}

	private static class SpadeOrder extends Card.Order {
		private int getValue(Card card) {
			if (card.equals(Card.QUEEN_SPADES))
				return 2;
			else if (card.equals(Card.KING_SPADES) || card.equals(Card.ACE_SPADES))
				return 1;
			else
				return 0;
		}
	
		protected Sort.Action compare(Card one, Card two) {
			if (getValue(one) > getValue(two))
				return Sort.Action.LEAVE;
			else if (getValue(two) > getValue(one))
				return Sort.Action.SWAP;
			else
				return Sort.Action.EQUAL;
		}
	}
	
	protected void initializeEvent(Intelligence.Environment environment) {
//		test.hearts.Security.openFile();
//		log("initialize called");
	}
	
	public CardArray getPass(int numberToPass) {
		CardArray cards = getCards().sort(DESCENDING);
		CardArray rv = new CardArray();
		for (int i=0; i<numberToPass; i++) {
			rv.add(cards.get(i));
		}
		return rv;
	}
	
	private Card getHighest() {
		return getLegalPlays().sort(DESCENDING).get(0);
	}
	
	private Card getLowest() {
		return getLegalPlays().sort(ASCENDING).get(0);
	}
	
	private Card getHighestWithoutTaking() {
		CardArray cards = getLegalPlays().sort(DESCENDING);
		for (int i=0; i<cards.getLength(); i++) {
			if (isHigher(getCurrentTrickWinner(), cards.get(i)))
				return cards.get(i);
		}
		return null;
	}
	
	private Card tryToPlayUnder() {
		if (getHighestWithoutTaking() != null)
			return getHighestWithoutTaking();
		return getLowest();
	}
	
	private boolean isFirstTrick() {
		return getHand().getPreviousTricks().length == 0;
	}
	
	private boolean isNextPlayLead() {
		return getTrick().getCards().isEmpty();
	}
	
	private boolean isNextPlayLast() {
		return getTrick().getCards().getLength() == players() - 1;
	}
	
	private int getCurrentTrickValue() {
		return getValue(getTrick().getCards());
	}
	
	private boolean isVoidedInTrickSuit() {
		return getCards().get( getTrick().getCards().get(0).getSuit() ).getLength() == 0;
	}
	
	private Card getCurrentTrickWinner() {
		if ( getTrick().getCards().isEmpty() ) {
			return null;
		} else {
			Card led = getTrick().getCards().get(0);
			Card leader = led;
			for (int i=1; i<getTrick().getCards().getLength(); i++) {
				if (isSameSuit(led, getTrick().getCards().get(i))) {
					if (isHigher(getTrick().getCards().get(i), leader)) {
						leader = getTrick().getCards().get(i);
					}
				}
			}
			return leader;
		}
	}
	
	private Card getHighestNoQueen() {
		Card rv = getHighest();
		if (rv.equals(Card.QUEEN_SPADES)) {
			log("highest queen");
			if (!getCurrentTrickWinner().getSuit().equals(Suit.SPADES)) {
				log("Can drop queen");
				return Card.QUEEN_SPADES;
			}
			if (getLegalPlays().getLength() > 1) {
				log("have alternative");
				return getLegalPlays().sort(DESCENDING).get(1);
			} else {
				log("no alternative");
				return Card.QUEEN_SPADES;
			}
		} else {
			return rv;
		}
	}

	public Card getPlay() {
		if (isFirstTrick()) {
			log("1st trick");
			return getHighest();
		} else if (isNextPlayLead()) {
			log("next play lead; get lowest");
			return getLowest();
		} else if (isNextPlayLast()) {
			log("last play -- is trick good?");
			if (getCurrentTrickValue() <= 0) {
				log("yes");
				//	if there are no points, take it
				if (
					getLegalPlays().contains(Card.JACK_DIAMONDS) 
					&& getCurrentTrickWinner().getSuit().equals(Suit.DIAMONDS)
					&& isHigher(Card.JACK_DIAMONDS, getCurrentTrickWinner())
				) {
					log("should take with jack");
					return Card.JACK_DIAMONDS;
				} else {
					log("get highest");
					return getHighestNoQueen();
				}
			} else {
				log("no; play highest without taking");
				//	otherwise play below it if possible
				if (getHighestWithoutTaking() != null) {
					return getHighestWithoutTaking();
				} else {
					//	if you're going to take it, take it big
					return getHighestNoQueen();
				}
			}
		} else if (isVoidedInTrickSuit()) {
			log("voided -- play highest");
			return getHighest();
		} else {
			log("not voided -- play under");
			if (getHighestWithoutTaking() != null)
				return getHighestWithoutTaking();
			else
				return getLowest();
		}
	}
	
	private static class SuitOrder extends Card.Order {
		private int getSuitValue(Suit suit) {
			if (suit.equals(Suit.SPADES))
				return 4;
			else if (suit.equals(Suit.HEARTS))
				return 3;
			else if (suit.equals(Suit.CLUBS))
				return 2;
			else if (suit.equals(Suit.DIAMONDS))
				return 1;
			else {
				Debug.unreachable("Unrecognized suit: " + suit.getName());
				return 0;
			}
		}
	
		public Sort.Action compare(Card one, Card two) {
			if (getSuitValue(one.getSuit()) > getSuitValue(two.getSuit()))
				return Sort.Action.LEAVE;
			else if (getSuitValue(two.getSuit()) > getSuitValue(one.getSuit()))
				return Sort.Action.SWAP;
			else
				return Sort.Action.EQUAL;
		}
	}
}